from typing import List

from pylic.__version__ import version
from pylic.cli.commands.command import Command
from pylic.cli.console_writer import END_STYLE, LABEL, console_writer


class VersionCommand(Command):
    targets = ["-V", "--version"]
    token = "version"

    def handle(self, options: List[str]) -> int:
        console_writer.line(f"Pylic version {LABEL}{version}{END_STYLE}")
        return 0



from typing import List

from pylic.__version__ import version
from pylic.cli.commands.command import Command
from pylic.cli.console_writer import BOLD, END_STYLE, LABEL, UNDERLINE, console_writer


class HelpCommand(Command):
    targets = ["-h", "--help"]
    token = "help"

    def handle(self, options: List[str]) -> int:
        console_writer.line(f"Pylic version {LABEL}{version}{END_STYLE}\n")
        console_writer.line(f'{self._header("USAGE")}')
        console_writer.line(f"  {UNDERLINE}pylic{END_STYLE} [-h] [-V] <command>\n")
        console_writer.line(f'{self._header("ARGUMENTS")}')
        console_writer.line(f"  {LABEL}<command>{END_STYLE}\t\tThe command to execute\n")
        console_writer.line(f'{self._header("GLOBAL OPTIONS")}')
        console_writer.line(f"  {LABEL}-h{END_STYLE} (--help)\t\tDisplay this or a commands help message")
        console_writer.line(f"  {LABEL}-V{END_STYLE} (--version)\tDisplay this application version\n")
        console_writer.line(f'{self._header("AVAILABLE COMMANDS")}')
        console_writer.line(f"  {LABEL}check{END_STYLE}\t\t\tChecks all installed licenses")
        console_writer.line(f"  {LABEL}list{END_STYLE}\t\t\tLists all installed packages and their corresponding license\n")
        return 1

    def _header(self, text: str) -> str:
        return f"{BOLD}{text}{END_STYLE}"



from typing import List

from pylic.cli.commands.command import Command
from pylic.cli.console_writer import BLUE, BOLD, END_STYLE, LABEL, UNDERLINE, WARNING, console_writer
from pylic.license_checker import LicenseChecker
from pylic.licenses import read_all_installed_licenses_metadata
from pylic.toml import read_config


class CheckCommand(Command):
    targets = ["check"]
    token = "check"

    def handle(self, options: List[str]) -> int:
        if "help" in options:
            self._show_help()
            return 1

        config = read_config()
        installed_licenses = read_all_installed_licenses_metadata()
        checker = LicenseChecker(config, installed_licenses)

        unnecessary_safe_licenses = checker.get_unnecessary_safe_licenses()
        unnecessary_unsafe_packages = checker.get_unnecessary_unsafe_packages()
        unnecessary_ignore_packages = checker.get_unnecessary_ignore_packages()
        bad_unsafe_packages = checker.get_bad_unsafe_packages()
        missing_unsafe_packages = checker.get_missing_unsafe_packages()
        unsafe_licenses = checker.get_unsafe_licenses()

        if any(
            [
                unnecessary_safe_licenses,
                unnecessary_unsafe_packages,
                unnecessary_ignore_packages,
                bad_unsafe_packages,
                missing_unsafe_packages,
                unsafe_licenses.found,
            ]
        ):
            if len(unnecessary_safe_licenses) > 0:
                console_writer.line("Unnecessary safe licenses listed which are not used by any installed package:")
                for unnecessary_safe_license in unnecessary_safe_licenses:
                    console_writer.line(f"  {WARNING}{unnecessary_safe_license}{END_STYLE}")

            if len(unnecessary_unsafe_packages) > 0:
                console_writer.line("Unsafe packages listed which are not installed:")
                for unnecessary_unsafe_package in unnecessary_unsafe_packages:
                    console_writer.line(f"  {WARNING}{unnecessary_unsafe_package}{END_STYLE}")

            if len(unnecessary_ignore_packages) > 0:
                console_writer.line("Ignore packages listed which are not installed:")
                for unnecessary_ignore_package in unnecessary_ignore_packages:
                    console_writer.line(f"  {WARNING}{unnecessary_ignore_package}{END_STYLE}")

            if len(bad_unsafe_packages) > 0:
                console_writer.line("Found unsafe packages with a known license. Instead allow these licenses explicitly:")
                for bad_package in bad_unsafe_packages:
                    console_writer.line(
                        (
                            f"  {WARNING}{bad_package['package']}{END_STYLE} {LABEL}({bad_package['version']}{END_STYLE}): "
                            f"{BLUE}{bad_package['license']}{END_STYLE}"
                        )
                    )

            if len(missing_unsafe_packages) > 0:
                console_writer.line("Found unsafe packages:")
                for missing_unsafe_package in missing_unsafe_packages:
                    console_writer.line(
                        (
                            f"  {WARNING}{missing_unsafe_package['package']}{END_STYLE} "
                            f"{LABEL}({missing_unsafe_package['version']}){END_STYLE}"
                        )
                    )

            if len(unsafe_licenses.found) > 0:
                console_writer.line("Found unsafe licenses:")
                for bad_license in unsafe_licenses.found:
                    console_writer.line(
                        (
                            f"  {BLUE}{bad_license['package']}{END_STYLE} {LABEL}({bad_license['version']}){END_STYLE}: "
                            f"{WARNING}{bad_license['license']}{END_STYLE}"
                        )
                    )

            return 1

        if len(unsafe_licenses.ignored) > 0:
            console_writer.line("Ignored packages with unsafe licenses:")
            for bad_license in unsafe_licenses.ignored:
                console_writer.line(
                    (
                        f"  {BLUE}{bad_license['package']}{END_STYLE} {LABEL}({bad_license['version']}){END_STYLE}: "
                        f"{WARNING}{bad_license['license']}{END_STYLE}"
                    )
                )

        console_writer.write_all_licenses_ok()
        return 0

    def _show_help(self) -> None:
        console_writer.line(f"{BOLD}USAGE{END_STYLE}")
        console_writer.line(f"  {UNDERLINE}pylic{END_STYLE} {UNDERLINE}check{END_STYLE} [-h]\n")
        console_writer.line(f"{BOLD}GLOBAL OPTIONS{END_STYLE}")
        console_writer.line(f"  {LABEL}-h{END_STYLE} (--help)\tDisplay this help message\n")
        console_writer.line(f"{BOLD}DESCRIPTION{END_STYLE}")
        console_writer.line("  Checks all installed licenses against the configuaration provided in the [tool.pylic]")
        console_writer.line("  section of the pyproject.toml file.\n")
        console_writer.line(f"    - {BOLD}safe_licenses{END_STYLE}: Licenses considered to be valid.")
        console_writer.line(f"    - {BOLD}unsafe_packages{END_STYLE}: Packages without a license yet to be considered valid.")
        console_writer.line(f"    - {BOLD}ignore_packages{END_STYLE}: Packages that are not considered at all.")



from dataclasses import dataclass
from typing import Dict, List, Optional

from pylic.toml import AppConfig


@dataclass
class UnsafeLicenses:
    found: List[Dict]
    ignored: List[Dict]


class LicenseChecker:
    def __init__(
        self,
        config: Optional[AppConfig] = None,
        installed_licenses: Optional[List[Dict]] = None,
    ) -> None:
        self.config = config or AppConfig()
        self.installed_licenses = installed_licenses or []

    def get_unnecessary_safe_licenses(self) -> List[str]:
        installed_license_names = [license_info["license"].lower() for license_info in self.installed_licenses]

        unnecessary_safe_licenses = []
        lower_safe_licenses = [safe_license.lower() for safe_license in self.config.safe_licenses]

        for index, safe_license in enumerate(lower_safe_licenses):
            if safe_license not in installed_license_names:
                unnecessary_safe_licenses.append(self.config.safe_licenses[index])

        return unnecessary_safe_licenses

    def get_unnecessary_unsafe_packages(self) -> List[str]:
        installed_package_names = [license_info["package"].lower() for license_info in self.installed_licenses]

        unnecessary_unsafe_packages = []
        lower_unsafe_packages = [unsafe_package.lower() for unsafe_package in self.config.unsafe_packages]

        for index, unsafe_package in enumerate(lower_unsafe_packages):
            if unsafe_package not in installed_package_names:
                unnecessary_unsafe_packages.append(self.config.unsafe_packages[index])

        return unnecessary_unsafe_packages

    def get_unnecessary_ignore_packages(self) -> List[str]:
        installed_package_names = [license_info["package"].lower() for license_info in self.installed_licenses]

        unnecessary_ignore_packages = []
        lower_ignore_packages = [ignore_package.lower() for ignore_package in self.config.ignore_packages]

        for index, ignore_package in enumerate(lower_ignore_packages):
            if ignore_package not in installed_package_names:
                unnecessary_ignore_packages.append(self.config.ignore_packages[index])

        return unnecessary_ignore_packages

    def get_bad_unsafe_packages(self) -> List[Dict]:
        bad_unsafe_packages: List[Dict] = []

        for license_info in self.installed_licenses:
            license = license_info["license"]
            package = license_info["package"]

            if package in self.config.unsafe_packages and license.lower() != "unknown":
                bad_unsafe_packages.append({"license": license, "package": package, "version": license_info["version"]})

        return bad_unsafe_packages

    def get_missing_unsafe_packages(self) -> List[Dict]:
        missing_unsafe_packages: List[Dict] = []

        for license_info in self.installed_licenses:
            license = license_info["license"]
            package = license_info["package"]

            if license.lower() == "unknown" and package not in self.config.unsafe_packages:
                missing_unsafe_packages.append({"package": package, "version": license_info["version"]})

        return missing_unsafe_packages

    def get_unsafe_licenses(self) -> UnsafeLicenses:
        found_licenses: List[Dict] = []
        ignored_licenses: List[Dict] = []

        lower_safe_licenses = [safe_license.lower() for safe_license in self.config.safe_licenses]

        for license_info in self.installed_licenses:
            license = license_info["license"]

            if license.lower() == "unknown" or license.lower() in lower_safe_licenses:
                continue

            datum = {"license": license, "package": license_info["package"], "version": license_info["version"]}
            if license_info["package"] in self.config.ignore_packages:
                ignored_licenses.append(datum)
            else:
                found_licenses.append(datum)

        return UnsafeLicenses(found=found_licenses, ignored=ignored_licenses)



from typing import List, NamedTuple

from pylic.cli.commands.command import Command
from pylic.cli.console_writer import console_writer


class Program(NamedTuple):
    command: Command
    options: List[str]


class ConsoleReader:
    def __init__(self, commands: List[Command]) -> None:
        self.target_to_token = {target: command for command in commands for target in command.targets}
        self.token_to_command = {command.token: command for command in commands}
        self.help_command = next(filter(lambda command: command.token == "help", commands))

    def get_program(self, input: List[str]) -> Program:
        tokens = self._tokenize_input(input)
        return self._parse_tokens(tokens)

    def _tokenize_input(self, raw_strings: List[str]) -> List[str]:
        tokens = []
        for raw in raw_strings:
            if raw.startswith("-"):
                candidate = raw[:2]  # single dash options only allow a single char
                if raw.startswith("--"):
                    candidate = raw
                command = self.target_to_token.get(candidate)
                if command is None:
                    console_writer.write_no_such_option(candidate)
                    exit(1)
                tokens.append(command.token)
                continue

            command = self.target_to_token.get(raw)
            if command is None:
                console_writer.write_no_such_command(raw)
                exit(1)
            tokens.append(command.token)

        return tokens

    def _parse_tokens(self, tokens: List[str]) -> Program:
        if len(tokens) == 0:
            return Program(self.help_command, [])

        return Program(self.token_to_command[tokens[0]], tokens[1:])



import sys
from typing import Dict, List

if sys.version_info[0] >= 3 and sys.version_info[1] >= 8:
    from importlib.metadata import Distribution, distributions
else:
    from importlib_metadata import Distribution, distributions


def read_all_installed_licenses_metadata() -> List[Dict]:
    installed_distributions = distributions()

    installed_licenses: List[Dict] = []
    for distribution in installed_distributions:
        license_string = _read_license_from_classifier(distribution)
        if license_string == "unknown":
            license_string = _read_license_from_metadata(distribution)
        if license_string == "OSI Approved":
            license_string = _read_license_from_metadata(distribution, fallback="OSI Approved")

        installed_licenses.append(
            {
                "license": license_string,
                "package": distribution.metadata["Name"],
                "version": distribution.metadata["Version"],
            }
        )

    return installed_licenses


def _read_license_from_classifier(distribution: Distribution) -> str:
    for key, content in distribution.metadata.items():
        if key == "Classifier":
            parts = [part.strip() for part in content.split("::")]
            if parts[0] == "License":
                return parts[-1]  # type: ignore[no-any-return]

    return "unknown"


def _read_license_from_metadata(distribution: Distribution, fallback: str = "unknown") -> str:
    return distribution.metadata.get("License", fallback)



import sys
from typing import Optional, TextIO

LABEL = "\033[96m"
SUCCESS = "\033[92m"
BLUE = "\033[94m"
WARNING = "\033[93m"
ERROR = "\033[91m"
BOLD = "\033[1m"
UNDERLINE = "\033[4m"
END_STYLE = "\033[0m"


class ConsoleWriter:
    def __init__(self, stream: Optional[TextIO] = None) -> None:
        self.stream = stream or sys.stdout

    def write_no_such_command(self, command: str) -> None:
        self.line(f'{ERROR}{BOLD}The command "{command}" is not available.{END_STYLE}')

    def write_no_such_option(self, option: str) -> None:
        self.line(f'{ERROR}{BOLD}The option "{option}" is not available.{END_STYLE}')

    def write_all_licenses_ok(self) -> None:
        self.line(f"{WARNING}{BOLD}✨{END_STYLE} {SUCCESS}All licenses ok{END_STYLE} {WARNING}{BOLD}✨{END_STYLE}")

    def line(self, line: str) -> None:
        self.stream.write(f"{line}\n")
        self.stream.flush()


console_writer = ConsoleWriter()



from dataclasses import dataclass, field
from typing import Any, List, MutableMapping

import toml


@dataclass
class AppConfig:
    safe_licenses: List[str] = field(default_factory=list)
    unsafe_packages: List[str] = field(default_factory=list)
    ignore_packages: List[str] = field(default_factory=list)


def read_config(filename: str = "pyproject.toml") -> AppConfig:
    project_config = _read_pyproject_file(filename)
    pylic_config = project_config.get("tool", {}).get("pylic", {})
    safe_licenses: List[str] = pylic_config.get("safe_licenses", [])

    if "unknown" in [safe_license.lower() for safe_license in safe_licenses]:
        raise ValueError("'unknown' can't be an safe license. Whitelist the corresponding packages instead.")

    unsafe_packages: List[str] = pylic_config.get("unsafe_packages", [])
    ignore_packages: List[str] = pylic_config.get("ignore_packages", [])

    return AppConfig(
        safe_licenses=safe_licenses,
        unsafe_packages=unsafe_packages,
        ignore_packages=ignore_packages,
    )


def _read_pyproject_file(filename: str) -> MutableMapping[str, Any]:
    with open(filename, "r") as pyproject_file:
        try:
            return toml.load(pyproject_file)
        except Exception as exception:
            raise exception



from typing import List

from pylic.cli.commands.command import Command
from pylic.cli.console_writer import BLUE, BOLD, END_STYLE, LABEL, UNDERLINE, console_writer
from pylic.licenses import read_all_installed_licenses_metadata


class ListCommand(Command):
    targets = ["list"]
    token = "list"

    def handle(self, options: List[str]) -> int:
        if "help" in options:
            self._show_help()
            return 1

        installed_licenses = read_all_installed_licenses_metadata()
        unsorted = {
            installed["package"]: {"version": installed["version"], "license": installed["license"]} for installed in installed_licenses
        }
        for package, rest in sorted(unsorted.items(), key=lambda k: k[0].lower()):  # type:ignore
            console_writer.line(f"{BLUE}{package}{END_STYLE} {LABEL}({rest['version']}){END_STYLE}: {rest['license']}")
        return 0

    def _show_help(self) -> None:
        console_writer.line(f"{BOLD}USAGE{END_STYLE}")
        console_writer.line(f"  {UNDERLINE}pylic{END_STYLE} {UNDERLINE}list{END_STYLE} [-h]\n")
        console_writer.line(f"{BOLD}GLOBAL OPTIONS{END_STYLE}")
        console_writer.line(f"  {LABEL}-h{END_STYLE} (--help)\tDisplay this help message\n")
        console_writer.line(f"{BOLD}DESCRIPTION{END_STYLE}")
        console_writer.line("  Lists all installed packages and their corresponding license.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pylic', 'pylic.cli', 'pylic.cli.commands']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=4.12.0,<5.0.0', 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['pylic = pylic.cli.app:main']}

setup_kwargs = {
    'name': 'pylic',
    'version': '3.1.1',
    'description': 'A Python license checker',
    'long_description': '# pylic - Python license checker [![GitHub license](https://img.shields.io/badge/license-MIT-blue.svg)](https://github.com/sandrochuber/pylic/blob/main/LICENSE) [![PyPI version](https://badge.fury.io/py/pylic.svg)](https://badge.fury.io/py/pylic/) [![Codecov](https://codecov.io/gh/ubersan/pylic//branch/main/graph/badge.svg)](https://codecov.io/gh/ubersan/pylic/)\n\nReads pylic configuration in `pyproject.toml` and checks licenses of installed packages recursively.\n\nPrinciples:\n- Every license has to be allowed explicitly (case-insensitive comparison).\n- All installed packages without a license are considered unsafe and have to be listed as such.\n\n> Only installed packages are checked for licenses. Packages/dependencies listed in `pyproject.toml` are ignored.\n\n## Installation\n\n```sh\npip install pylic\n```\n\n## Configuration\n\n`pylic` needs be run in the directory where your `pyproject.toml` file is located. You can configure\n- `safe_licenses`: All licenses you consider safe for usage. The string comparison is case-insensitive.\n- `unsafe_packages`: If you rely on a package that does not come with a license you have to explicitly list it as such.\n- `ignore_packages`: Packages that will not be reported as unsafe even if they use a license not listed as safe. This is useful in case an existing projects want to start integrating `pylic`, but are still using unsafe licenses. This enables first to ignore these packages temporarely, while they\'re being replaced, second to already validate newly added or updated packages against the safe license set and third to integrate `pylic` frictionless into CI/CD from the get go.\n\n```toml\n[tool.pylic]\nsafe_licenses = [\n    "Apache Software License",\n    "Apache License 2.0",\n    "MIT License",\n    "Python Software Foundation License",\n    "Mozilla Public License 2.0 (MPL 2.0)",\n]\nunsafe_packages = [\n    "unlicensedPackage",\n]\nignore_packages = [\n    "ignoredPackage",\n]\n```\n\n## Commands\n\n`pylic` provides the following commands (also see `pylic help`):\n- `check`: Checks all installed licenses.\n- `list`: Lists all installed packages and their corresponding license.\n\n## Usage Example\n\nCreate a venv to start with a clean ground and activate it\n\n```sh\npython -m venv .venv\nsource .venv/bin/activate\n```\n\nInstall `pylic` and create an empty `pyproject.toml`\n\n```sh\npip install pylic\ntouch pyproject.toml\n```\n\nInstall all your dependencies\n\n```sh\npip install <packageA> <packageB>\n```\n\nRun pylic\n\n```sh\npylic check\n```\n\nThe output will be similar to\n\n```sh\nFound unsafe packages:\n  pkg_resources (0.0.0)\nFound unsafe licenses:\n  pip (18.1): MIT License\n  zipp (3.4.1): MIT License\n  toml (0.10.2): MIT License\n  pylic (1.2.0): MIT License\n  setuptools (40.8.0): MIT License\n  typing-extensions (3.7.4.3): Python Software Foundation License\n  importlib-metadata (3.9.0): Apache Software License\n```\n\nThe return code of `pylic` is in this case non-zero due to unsafe licenses. This allows usage of pylic in CI.\n\n```sh\necho $? # prints 1\n```\n\nAs these licenses and packages are all ok we can configure `pylic` accordingly\n\n```sh\ncat <<EOT >> pyproject.toml\n[tool.pylic]\nsafe_licenses = ["Apache Software License", "MIT License", "Python Software Foundation License"]\nunsafe_packages = ["pkg_resources"]\nEOT\n```\n\nAfter rerunning `pylic check` the output now reveals a successful validation\n\n```sh\n✨ All licenses ok ✨\n```\n\nAlso the return code now signals that all is good\n\n```sh\necho $? # prints 0\n```\n\nUse `pylic list` to list all installed packages and their corresponding licenses.\n\n## Development\n\nRequired tools:\n- Poetry (https://python-poetry.org/)\n\nRun `poetry install` to install all necessary dependencies. Checkout the `[tool.taskipy.tasks]` (see [taskipy](https://github.com/illBeRoy/taskipy)) section in the `pyproject.toml` file for utility tasks. You can run these with `poetry run task <task>`.\n\nCreating a new release is as simple as:\n- Update `version` in the pyproject.toml and the `__version__.py` file.\n- `poetry run task release`.\n',
    'author': 'Sandro Huber',
    'author_email': 'sandrochuber@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ubersan/pylic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys

from pylic.cli.commands.check import CheckCommand
from pylic.cli.commands.help import HelpCommand
from pylic.cli.commands.list import ListCommand
from pylic.cli.commands.version import VersionCommand
from pylic.cli.console_reader import ConsoleReader


def main() -> None:
    sys.argv.pop(0)

    program = ConsoleReader(
        commands=[
            CheckCommand(),
            ListCommand(),
            HelpCommand(),
            VersionCommand(),
        ]
    ).get_program(sys.argv)

    status_code = program.command.handle(program.options)
    exit(status_code)


if __name__ == "__main__":
    main()
